from unittest import TestCase

from mock import Mock, patch

from django_elasticsearch_dsl.documents import DocType
from django_elasticsearch_dsl.registries import registry
from django_elasticsearch_dsl.signals import post_index

from .models import Car


class PostIndexSignalTestCase(TestCase):

    @patch('django_elasticsearch_dsl.documents.DocType._get_actions')
    @patch('django_elasticsearch_dsl.documents.bulk')
    def test_post_index_signal_sent(self, bulk, get_actions):

        @registry.register_document
        class CarDocument(DocType):
            class Django:
                fields = ['name']
                model = Car

        bulk.return_value = (1, [])

        # register a mock signal receiver
        mock_receiver = Mock()
        post_index.connect(mock_receiver)

        doc = CarDocument()
        car = Car(
            pk=51,
            name="Type 57"
        )
        doc.update(car)

        bulk.assert_called_once()

        mock_receiver.assert_called_once_with(
            signal=post_index,
            sender=CarDocument,
            instance=doc,
            actions=get_actions(),
            response=(1, [])
        )



from mock import DEFAULT, Mock, patch
from unittest import TestCase

from django.core.management.base import CommandError
from django.core.management import call_command
from six import StringIO

from django_elasticsearch_dsl import Index
from django_elasticsearch_dsl.management.commands.search_index import Command
from django_elasticsearch_dsl.registries import DocumentRegistry

from .fixtures import WithFixturesMixin


class SearchIndexTestCase(WithFixturesMixin, TestCase):
    def _mock_setup(self):
        # Mock Patch object
        patch_registry = patch(
            'django_elasticsearch_dsl.management.commands.search_index.registry', self.registry)

        patch_registry.start()

        methods = ['delete', 'create']
        for index in [self.index_a, self.index_b]:
            for method in methods:
                obj_patch = patch.object(index, method)
                obj_patch.start()

        self.addCleanup(patch.stopall)

    def setUp(self):
        self.out = StringIO()
        self.registry = DocumentRegistry()
        self.index_a = Index('foo')
        self.index_b = Index('bar')

        self.doc_a1_qs = Mock()
        self.doc_a1 = self._generate_doc_mock(
            self.ModelA, self.index_a, self.doc_a1_qs
        )

        self.doc_a2_qs = Mock()
        self.doc_a2 = self._generate_doc_mock(
            self.ModelA, self.index_a, self.doc_a2_qs
        )

        self.doc_b1_qs = Mock()
        self.doc_b1 = self._generate_doc_mock(
            self.ModelB, self.index_a, self.doc_b1_qs
        )

        self.doc_c1_qs = Mock()
        self.doc_c1 = self._generate_doc_mock(
            self.ModelC, self.index_b, self.doc_c1_qs
        )

        self._mock_setup()

    def test_get_models(self):
        cmd = Command()
        self.assertEqual(
            cmd._get_models(['foo']),
            set([self.ModelA, self.ModelB])
        )

        self.assertEqual(
            cmd._get_models(['foo', 'bar.ModelC']),
            set([self.ModelA, self.ModelB, self.ModelC])
        )

        self.assertEqual(
            cmd._get_models([]),
            set([self.ModelA, self.ModelB, self.ModelC])
        )
        with self.assertRaises(CommandError):
            cmd._get_models(['unknown'])

    def test_no_action_error(self):
        cmd = Command()
        with self.assertRaises(CommandError):
            cmd.handle(action="")

    def test_delete_foo_index(self):

        with patch(
            'django_elasticsearch_dsl.management.commands.search_index.input',
            Mock(return_value="y")
        ):
            call_command('search_index', stdout=self.out,
                         action='delete', models=['foo'])
            self.index_a.delete.assert_called_once()
            self.assertFalse(self.index_b.delete.called)

    def test_force_delete_all_indices(self):

        call_command('search_index', stdout=self.out,
                     action='delete', force=True)
        self.index_a.delete.assert_called_once()
        self.index_b.delete.assert_called_once()

    def test_force_delete_bar_model_c_index(self):
        call_command('search_index', stdout=self.out,
                     models=[self.ModelC._meta.label],
                     action='delete', force=True)
        self.index_b.delete.assert_called_once()
        self.assertFalse(self.index_a.delete.called)

    def test_create_all_indices(self):
        call_command('search_index', stdout=self.out, action='create')
        self.index_a.create.assert_called_once()
        self.index_b.create.assert_called_once()

    def test_populate_all_doc_type(self):
        call_command('search_index', stdout=self.out, action='populate')
        expected_kwargs = {'parallel': False, 'refresh': None}
        # One call for "Indexing NNN documents", one for indexing itself (via get_index_queryset).
        assert self.doc_a1.get_queryset.call_count == 2
        self.doc_a1.update.assert_called_once_with(self.doc_a1_qs.iterator(), **expected_kwargs)
        assert self.doc_a2.get_queryset.call_count == 2
        self.doc_a2.update.assert_called_once_with(self.doc_a2_qs.iterator(), **expected_kwargs)
        assert self.doc_b1.get_queryset.call_count == 2
        self.doc_b1.update.assert_called_once_with(self.doc_b1_qs.iterator(), **expected_kwargs)
        assert self.doc_c1.get_queryset.call_count == 2
        self.doc_c1.update.assert_called_once_with(self.doc_c1_qs.iterator(), **expected_kwargs)

    def test_populate_all_doc_type_refresh(self):
        call_command('search_index', stdout=self.out, action='populate', refresh=True)
        expected_kwargs = {'parallel': False, 'refresh': True}
        self.doc_a1.update.assert_called_once_with(self.doc_a1_qs.iterator(), **expected_kwargs)
        self.doc_a2.update.assert_called_once_with(self.doc_a2_qs.iterator(), **expected_kwargs)
        self.doc_b1.update.assert_called_once_with(self.doc_b1_qs.iterator(), **expected_kwargs)
        self.doc_c1.update.assert_called_once_with(self.doc_c1_qs.iterator(), **expected_kwargs)

    def test_rebuild_indices(self):

        with patch.multiple(
            Command, _create=DEFAULT, _delete=DEFAULT, _populate=DEFAULT
        ) as handles:
            handles['_delete'].return_value = True
            call_command('search_index', stdout=self.out, action='rebuild')
            handles['_delete'].assert_called()
            handles['_create'].assert_called()
            handles['_populate'].assert_called()

    def test_rebuild_indices_aborted(self):

        with patch.multiple(
            Command, _create=DEFAULT, _delete=DEFAULT, _populate=DEFAULT
        ) as handles:
            handles['_delete'].return_value = False
            call_command('search_index', stdout=self.out, action='rebuild')
            handles['_delete'].assert_called()
            handles['_create'].assert_not_called()
            handles['_populate'].assert_not_called()



from unittest import TestCase

import django
from django.db.models.fields.files import FieldFile
if django.VERSION < (4, 0):
    from django.utils.translation import ugettext_lazy as _
else:
    from django.utils.translation import gettext_lazy as _
from mock import Mock, NonCallableMock
from six import string_types

from django_elasticsearch_dsl.exceptions import VariableLookupError
from django_elasticsearch_dsl.fields import (BooleanField, ByteField, CompletionField, DEDField,
                                             DateField, DoubleField, FileField, FloatField,
                                             GeoPointField,
                                             GeoShapeField, IntegerField, IpField, KeywordField,
                                             ListField, LongField,
                                             NestedField, ObjectField, ScaledFloatField, ShortField, TextField
                                             )
from tests import ES_MAJOR_VERSION


class DEDFieldTestCase(TestCase):
    def test_attr_to_path(self):
        field = DEDField(attr='field')
        self.assertEqual(field._path, ['field'])

        field = DEDField(attr='obj.field')
        self.assertEqual(field._path, ['obj', 'field'])

    def test_get_value_from_instance_attr(self):
        field = DEDField(attr='attr1')
        instance = NonCallableMock(attr1="foo", attr2="bar")
        self.assertEqual(field.get_value_from_instance(instance), "foo")

    def test_get_value_from_instance_related_attr(self):
        field = DEDField(attr='related.attr1')
        instance = NonCallableMock(attr1="foo",
                                   related=NonCallableMock(attr1="bar"))
        self.assertEqual(field.get_value_from_instance(instance), "bar")

    def test_get_value_from_instance_callable(self):
        field = DEDField(attr='callable')
        instance = NonCallableMock(callable=Mock(return_value="bar"))
        self.assertEqual(field.get_value_from_instance(instance), "bar")

    def test_get_value_from_instance_related_callable(self):
        field = DEDField(attr='related.callable')
        instance = NonCallableMock(related=NonCallableMock(
            callable=Mock(return_value="bar"), attr1="foo"))
        self.assertEqual(field.get_value_from_instance(instance), "bar")

    def test_get_value_from_instance_with_unknown_attr(self):
        class Dummy:
            attr1 = "foo"

        field = DEDField(attr='attr2', required=True)
        self.assertRaises(
            VariableLookupError, field.get_value_from_instance, Dummy()
        )

    def test_get_value_from_none(self):
        field = DEDField(attr='related.none')
        instance = NonCallableMock(attr1="foo", related=None)
        self.assertEqual(field.get_value_from_instance(instance), None)

    def test_get_value_from_lazy_object(self):
        field = DEDField(attr='translation')
        instance = NonCallableMock(translation=_("foo"))
        self.assertIsInstance(
            field.get_value_from_instance(instance), string_types
        )
        self.assertEqual(field.get_value_from_instance(instance), "foo")


class ObjectFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = ObjectField(attr='person', properties={
            'first_name': TextField(analyzer='foo'),
            'last_name': TextField()
        })

        expected_type = 'string' if ES_MAJOR_VERSION == 2 else 'text'

        self.assertEqual({
            'type': 'object',
            'properties': {
                'first_name': {'type': expected_type, 'analyzer': 'foo'},
                'last_name': {'type': expected_type},
            }
        }, field.to_dict())

    def test_get_value_from_instance(self):
        field = ObjectField(attr='person', properties={
            'first_name': TextField(analyzer='foo'),
            'last_name': TextField()
        })

        instance = NonCallableMock(person=NonCallableMock(
            first_name='foo', last_name='bar'))

        self.assertEqual(field.get_value_from_instance(instance), {
            'first_name': "foo",
            'last_name': "bar",
        })

    def test_get_value_from_instance_with_partial_properties(self):
        field = ObjectField(
            attr='person',
            properties={
                'first_name': TextField(analyzer='foo')
            }
        )

        instance = NonCallableMock(
            person=NonCallableMock(first_name='foo', last_name='bar')
        )

        self.assertEqual(field.get_value_from_instance(instance), {
            'first_name': "foo"
        })

    def test_get_value_from_instance_without_properties(self):
        field = ObjectField(attr='person')

        instance = NonCallableMock(
            person={'first_name': 'foo', 'last_name': 'bar'}
        )

        self.assertEqual(field.get_value_from_instance(instance),
                         {
                             'first_name': "foo",
                             'last_name': "bar"
                         }
                         )

    def test_get_value_from_instance_with_inner_objectfield(self):
        field = ObjectField(attr='person', properties={
            'first_name': TextField(analyzer='foo'),
            'last_name': TextField(),
            'additional': ObjectField(properties={
                'age': IntegerField()
            })
        })

        instance = NonCallableMock(person=NonCallableMock(
            first_name="foo", last_name="bar",
            additional=NonCallableMock(age=12)
        ))

        self.assertEqual(field.get_value_from_instance(instance), {
            'first_name': "foo",
            'last_name': "bar",
            'additional': {'age': 12}
        })

    def test_get_value_from_instance_with_inner_objectfield_without_properties(self):
        field = ObjectField(
            attr='person',
            properties={
                'first_name': TextField(analyzer='foo'),
                'last_name': TextField(),
                'additional': ObjectField()
            }
        )

        instance = NonCallableMock(person=NonCallableMock(
            first_name="foo",
            last_name="bar",
            additional={'age': 12}
        ))

        self.assertEqual(field.get_value_from_instance(instance),
                         {
                             'first_name': "foo",
                             'last_name': "bar",
                             'additional': {'age': 12}
                         }
                         )

    def test_get_value_from_instance_with_none_inner_objectfield(self):
        field = ObjectField(attr='person', properties={
            'first_name': TextField(analyzer='foo'),
            'last_name': TextField(),
            'additional': ObjectField(properties={
                'age': IntegerField()
            })
        })

        instance = NonCallableMock(person=NonCallableMock(
            first_name="foo", last_name="bar",
            additional=None
        ))

        self.assertEqual(field.get_value_from_instance(instance), {
            'first_name': "foo",
            'last_name': "bar",
            'additional': {}
        })

    def test_get_value_from_iterable(self):
        field = ObjectField(attr='person', properties={
            'first_name': TextField(analyzer='foo'),
            'last_name': TextField()
        })

        instance = NonCallableMock(
            person=[
                NonCallableMock(
                    first_name="foo1", last_name="bar1"
                ),
                NonCallableMock(
                    first_name="foo2", last_name="bar2"
                )
            ]
        )

        self.assertEqual(field.get_value_from_instance(instance), [
            {
                'first_name': "foo1",
                'last_name': "bar1",
            },
            {
                'first_name': "foo2",
                'last_name': "bar2",
            }
        ])

    def test_get_value_from_iterable_without_properties(self):
        field = ObjectField(attr='person')

        instance = NonCallableMock(
            person=[
                {'first_name': "foo1", 'last_name': "bar1"},
                {'first_name': "foo2", 'last_name': "bar2"}
            ]
        )

        self.assertEqual(field.get_value_from_instance(instance),
                         [
                             {
                                 'first_name': "foo1",
                                 'last_name': "bar1",
                             },
                             {
                                 'first_name': "foo2",
                                 'last_name': "bar2",
                             }
                         ]
                         )


class NestedFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = NestedField(attr='person', properties={
            'first_name': TextField(analyzer='foo'),
            'last_name': TextField()
        })

        expected_type = 'string' if ES_MAJOR_VERSION == 2 else 'text'

        self.assertEqual({
            'type': 'nested',
            'properties': {
                'first_name': {'type': expected_type, 'analyzer': 'foo'},
                'last_name': {'type': expected_type},
            }
        }, field.to_dict())


class BooleanFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = BooleanField()

        self.assertEqual({
            'type': 'boolean',
        }, field.to_dict())


class DateFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = DateField()

        self.assertEqual({
            'type': 'date',
        }, field.to_dict())


class CompletionFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = CompletionField()

        self.assertEqual({
            'type': 'completion',
        }, field.to_dict())


class GeoPointFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = GeoPointField()

        self.assertEqual({
            'type': 'geo_point',
        }, field.to_dict())


class GeoShapeFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = GeoShapeField()

        self.assertEqual({
            'type': 'geo_shape'
        }, field.to_dict())


class ByteFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = ByteField()

        self.assertEqual({
            'type': 'byte',
        }, field.to_dict())


class LongFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = LongField()

        self.assertEqual({
            'type': 'long',
        }, field.to_dict())


class DoubleFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = DoubleField()

        self.assertEqual({
            'type': 'double',
        }, field.to_dict())


class FloatFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = FloatField()

        self.assertEqual({
            'type': 'float',
        }, field.to_dict())


class ScaledFloatFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = ScaledFloatField(scaling_factor=100)

        self.assertEqual({
            'type': 'scaled_float',
            'scaling_factor': 100,
        }, field.to_dict())


class IpFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = IpField()

        self.assertEqual({
            'type': 'ip',
        }, field.to_dict())


class ListFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = ListField(IntegerField(attr='foo.bar'))
        self.assertEqual({
            'type': 'integer',
        }, field.to_dict())

    def test_get_value_from_instance(self):
        instance = NonCallableMock(
            foo=NonCallableMock(bar=["alpha", "beta", "gamma"])
        )
        field = ListField(TextField(attr='foo.bar'))
        self.assertEqual(
            field.get_value_from_instance(instance), instance.foo.bar)


class ShortFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = ShortField()

        self.assertEqual({
            'type': 'short',
        }, field.to_dict())


class FileFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = FileField()

        expected_type = 'string' if ES_MAJOR_VERSION == 2 else 'text'

        self.assertEqual({
            'type': expected_type,
        }, field.to_dict())

    def test_get_value_from_instance(self):
        field = FileField(attr='file')

        instance = NonCallableMock(
            file=NonCallableMock(spec=FieldFile, url='myfile.pdf'),
        )
        self.assertEqual(
            field.get_value_from_instance(instance), 'myfile.pdf'
        )

        field = FileField(attr='related.attr1')
        instance = NonCallableMock(
            attr1="foo", related=NonCallableMock(attr1="bar")
        )
        self.assertEqual(field.get_value_from_instance(instance), 'bar')


class TextFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = TextField()

        expected_type = 'string' if ES_MAJOR_VERSION == 2 else 'text'

        self.assertEqual({
            'type': expected_type,
        }, field.to_dict())


class KeywordFieldTestCase(TestCase):
    def test_get_mapping(self):
        field = KeywordField()

        if ES_MAJOR_VERSION == 2:
            self.assertEqual({
                'type': 'string',
                'index': 'not_analyzed',
            }, field.to_dict())
        else:
            self.assertEqual({
                'type': 'keyword',
            }, field.to_dict())



from collections import defaultdict
from copy import deepcopy

from itertools import chain

from django.core.exceptions import ObjectDoesNotExist
from django.core.exceptions import ImproperlyConfigured
from elasticsearch_dsl import AttrDict
from six import itervalues, iterkeys, iteritems

from django_elasticsearch_dsl.exceptions import RedeclaredFieldError
from .apps import DEDConfig


class DocumentRegistry(object):
    """
    Registry of models classes to a set of Document classes.
    """
    def __init__(self):
        self._indices = defaultdict(set)
        self._models = defaultdict(set)
        self._related_models = defaultdict(set)

    def register(self, index, doc_class):
        """Register the model with the registry"""
        self._models[doc_class.django.model].add(doc_class)

        for related in doc_class.django.related_models:
            self._related_models[related].add(doc_class.django.model)

        for idx, docs in iteritems(self._indices):
            if index._name == idx._name:
                docs.add(doc_class)
                return

        self._indices[index].add(doc_class)

    def register_document(self, document):
        django_meta = getattr(document, 'Django')
        # Raise error if Django class can not be found
        if not django_meta:
            message = "You must declare the Django class inside {}".format(document.__name__)
            raise ImproperlyConfigured(message)

        # Keep all django related attribute in a django_attr AttrDict
        data = {'model': getattr(document.Django, 'model')}
        django_attr = AttrDict(data)

        if not django_attr.model:
            raise ImproperlyConfigured("You must specify the django model")

        # Add The model fields into elasticsearch mapping field
        model_field_names = getattr(document.Django, "fields", [])
        mapping_fields = document._doc_type.mapping.properties.properties.to_dict().keys()

        for field_name in model_field_names:
            if field_name in mapping_fields:
                raise RedeclaredFieldError(
                    "You cannot redeclare the field named '{}' on {}"
                    .format(field_name, document.__name__)
                )

            django_field = django_attr.model._meta.get_field(field_name)

            field_instance = document.to_field(field_name, django_field)
            document._doc_type.mapping.field(field_name, field_instance)

        django_attr.ignore_signals = getattr(django_meta, "ignore_signals", False)
        django_attr.auto_refresh = getattr(django_meta,
                                           "auto_refresh", DEDConfig.auto_refresh_enabled())
        django_attr.related_models = getattr(django_meta, "related_models", [])
        django_attr.queryset_pagination = getattr(django_meta, "queryset_pagination", None)

        # Add django attribute in the document class with all the django attribute
        setattr(document, 'django', django_attr)

        # Set the fields of the mappings
        fields = document._doc_type.mapping.properties.properties.to_dict()
        setattr(document, '_fields', fields)

        # Update settings of the document index
        default_index_settings = deepcopy(DEDConfig.default_index_settings())
        document._index.settings(**default_index_settings)

        # Register the document and index class to our registry
        self.register(index=document._index, doc_class=document)

        return document

    def _get_related_doc(self, instance):
        for model in self._related_models.get(instance.__class__, []):
            for doc in self._models[model]:
                if instance.__class__ in doc.django.related_models:
                    yield doc

    def update_related(self, instance, **kwargs):
        """
        Update docs that have related_models.
        """
        if not DEDConfig.autosync_enabled():
            return

        for doc in self._get_related_doc(instance):
            doc_instance = doc()
            try:
                related = doc_instance.get_instances_from_related(instance)
            except ObjectDoesNotExist:
                related = None

            if related is not None:
                doc_instance.update(related, **kwargs)

    def delete_related(self, instance, **kwargs):
        """
        Remove `instance` from related models.
        """
        if not DEDConfig.autosync_enabled():
            return

        for doc in self._get_related_doc(instance):
            doc_instance = doc(related_instance_to_ignore=instance)
            try:
                related = doc_instance.get_instances_from_related(instance)
            except ObjectDoesNotExist:
                related = None

            if related is not None:
                doc_instance.update(related, **kwargs)

    def update(self, instance, **kwargs):
        """
        Update all the elasticsearch documents attached to this model (if their
        ignore_signals flag allows it)
        """
        if not DEDConfig.autosync_enabled():
            return

        if instance.__class__ in self._models:
            for doc in self._models[instance.__class__]:
                if not doc.django.ignore_signals:
                    doc().update(instance, **kwargs)

    def delete(self, instance, **kwargs):
        """
        Delete all the elasticsearch documents attached to this model (if their
        ignore_signals flag allows it)
        """
        self.update(instance, action="delete", **kwargs)

    def get_documents(self, models=None):
        """
        Get all documents in the registry or the documents for a list of models
        """
        if models is not None:
            return set(chain.from_iterable(self._models[model] for model in models
                                           if model in self._models))
        return set(chain.from_iterable(itervalues(self._indices)))

    def get_models(self):
        """
        Get all models in the registry
        """
        return set(iterkeys(self._models))

    def get_indices(self, models=None):
        """
        Get all indices in the registry or the indices for a list of models
        """
        if models is not None:
            return set(
                indice for indice, docs in iteritems(self._indices)
                for doc in docs if doc.django.model in models
            )

        return set(iterkeys(self._indices))

    def __contains__(self, model):
        """
        Checks that model is in registry
        """
        return model in self._models or model in self._related_models


registry = DocumentRegistry()



from __future__ import unicode_literals

from collections import deque
from fnmatch import fnmatch
from functools import partial

from django import VERSION as DJANGO_VERSION
from django.db import models
from elasticsearch.helpers import bulk, parallel_bulk
from elasticsearch_dsl import Document as DSLDocument
from six import iteritems

from .exceptions import ModelFieldNotMappedError
from .fields import (
    BooleanField,
    DateField,
    DEDField,
    DoubleField,
    FileField,
    IntegerField,
    KeywordField,
    LongField,
    ShortField,
    TextField, TimeField,
)
from .search import Search
from .signals import post_index

model_field_class_to_field_class = {
    models.AutoField: IntegerField,
    models.BigAutoField: LongField,
    models.BigIntegerField: LongField,
    models.BooleanField: BooleanField,
    models.CharField: TextField,
    models.DateField: DateField,
    models.DateTimeField: DateField,
    models.DecimalField: DoubleField,
    models.EmailField: TextField,
    models.FileField: FileField,
    models.FilePathField: KeywordField,
    models.FloatField: DoubleField,
    models.ImageField: FileField,
    models.IntegerField: IntegerField,
    models.NullBooleanField: BooleanField,
    models.PositiveIntegerField: IntegerField,
    models.PositiveSmallIntegerField: ShortField,
    models.SlugField: KeywordField,
    models.SmallIntegerField: ShortField,
    models.TextField: TextField,
    models.TimeField: TimeField,
    models.URLField: TextField,
    models.UUIDField: KeywordField,
}

if DJANGO_VERSION >= (3.1,):
    model_field_class_to_field_class[models.PositiveBigIntegerField] = LongField


class DocType(DSLDocument):
    _prepared_fields = []

    def __init__(self, related_instance_to_ignore=None, **kwargs):
        super(DocType, self).__init__(**kwargs)
        self._related_instance_to_ignore = related_instance_to_ignore
        self._prepared_fields = self.init_prepare()

    def __eq__(self, other):
        return id(self) == id(other)

    def __hash__(self):
        return id(self)

    @classmethod
    def _matches(cls, hit):
        """
        Determine which index or indices in a pattern to be used in a hit.
        Overrides DSLDocument _matches function to match indices in a pattern,
        which is needed in case of using aliases. This is needed as the
        document class will not be used to deserialize the documents. The
        documents will have the index set to the concrete index, whereas the
        class refers to the alias.
        """
        return fnmatch(hit.get("_index", ""), cls._index._name + "*")

    @classmethod
    def search(cls, using=None, index=None):
        return Search(
            using=cls._get_using(using),
            index=cls._default_index(index),
            doc_type=[cls],
            model=cls.django.model
        )

    def get_queryset(self):
        """
        Return the queryset that should be indexed by this doc type.
        """
        return self.django.model._default_manager.all()

    def get_indexing_queryset(self):
        """
        Build queryset (iterator) for use by indexing.
        """
        qs = self.get_queryset()
        kwargs = {}
        if DJANGO_VERSION >= (2,) and self.django.queryset_pagination:
            kwargs = {'chunk_size': self.django.queryset_pagination}
        return qs.iterator(**kwargs)

    def init_prepare(self):
        """
        Initialise the data model preparers once here. Extracts the preparers
        from the model and generate a list of callables to avoid doing that
        work on every object instance over.
        """
        index_fields = getattr(self, '_fields', {})
        fields = []
        for name, field in iteritems(index_fields):
            if not isinstance(field, DEDField):
                continue

            if not field._path:
                field._path = [name]

            prep_func = getattr(self, 'prepare_%s_with_related' % name, None)
            if prep_func:
                fn = partial(prep_func, related_to_ignore=self._related_instance_to_ignore)
            else:
                prep_func = getattr(self, 'prepare_%s' % name, None)
                if prep_func:
                    fn = prep_func
                else:
                    fn = partial(field.get_value_from_instance, field_value_to_ignore=self._related_instance_to_ignore)

            fields.append((name, field, fn))

        return fields

    def prepare(self, instance):
        """
        Take a model instance, and turn it into a dict that can be serialized
        based on the fields defined on this DocType subclass
        """
        data = {
            name: prep_func(instance)
            for name, field, prep_func in self._prepared_fields
        }
        return data

    @classmethod
    def get_model_field_class_to_field_class(cls):
        """
        Returns dict of relationship from model field class to elasticsearch
        field class

        You may want to override this if you have model field class not included
        in model_field_class_to_field_class.
        """
        return model_field_class_to_field_class

    @classmethod
    def to_field(cls, field_name, model_field):
        """
        Returns the elasticsearch field instance appropriate for the model
        field class. This is a good place to hook into if you have more complex
        model field to ES field logic
        """
        try:
            return cls.get_model_field_class_to_field_class()[
                model_field.__class__](attr=field_name)
        except KeyError:
            raise ModelFieldNotMappedError(
                "Cannot convert model field {} "
                "to an Elasticsearch field!".format(field_name)
            )

    def bulk(self, actions, **kwargs):
        response = bulk(client=self._get_connection(), actions=actions, **kwargs)
        # send post index signal
        post_index.send(
            sender=self.__class__,
            instance=self,
            actions=actions,
            response=response
        )
        return response

    def parallel_bulk(self, actions, **kwargs):
        if self.django.queryset_pagination and 'chunk_size' not in kwargs:
            kwargs['chunk_size'] = self.django.queryset_pagination
        bulk_actions = parallel_bulk(client=self._get_connection(), actions=actions, **kwargs)
        # As the `parallel_bulk` is lazy, we need to get it into `deque` to run it instantly
        # See https://discuss.elastic.co/t/helpers-parallel-bulk-in-python-not-working/39498/2
        deque(bulk_actions, maxlen=0)
        # Fake return value to emulate bulk() since we don't have a result yet,
        # the result is currently not used upstream anyway.
        return (1, [])

    @classmethod
    def generate_id(cls, object_instance):
        """
        The default behavior is to use the Django object's pk (id) as the
        elasticseach index id (_id). If needed, this method can be overloaded
        to change this default behavior.
        """
        return object_instance.pk

    def _prepare_action(self, object_instance, action):
        return {
            '_op_type': action,
            '_index': self._index._name,
            '_id': self.generate_id(object_instance),
            '_source': (
                self.prepare(object_instance) if action != 'delete' else None
            ),
        }

    def _get_actions(self, object_list, action):
        for object_instance in object_list:
            if action == 'delete' or self.should_index_object(object_instance):
                yield self._prepare_action(object_instance, action)
    
    def get_actions(self, object_list, action):
        """
        Generate the elasticsearch payload.
        """
        return self._get_actions(object_list, action)


    def _bulk(self, *args, **kwargs):
        """Helper for switching between normal and parallel bulk operation"""
        parallel = kwargs.pop('parallel', False)
        if parallel:
            return self.parallel_bulk(*args, **kwargs)
        else:
            return self.bulk(*args, **kwargs)

    def should_index_object(self, obj):
        """
        Overwriting this method and returning a boolean value
        should determine whether the object should be indexed.
        """
        return True

    def update(self, thing, refresh=None, action='index', parallel=False, **kwargs):
        """
        Update each document in ES for a model, iterable of models or queryset
        """
        if refresh is not None:
            kwargs['refresh'] = refresh
        elif self.django.auto_refresh:
            kwargs['refresh'] = self.django.auto_refresh

        if isinstance(thing, models.Model):
            object_list = [thing]
        else:
            object_list = thing

        return self._bulk(
            self._get_actions(object_list, action),
            parallel=parallel,
            **kwargs
        )


# Alias of DocType. Need to remove DocType in 7.x
Document = DocType



from copy import deepcopy

from elasticsearch_dsl import Index as DSLIndex
from six import python_2_unicode_compatible

from .apps import DEDConfig
from .registries import registry


@python_2_unicode_compatible
class Index(DSLIndex):
    def __init__(self, *args, **kwargs):
        super(Index, self).__init__(*args, **kwargs)
        default_index_settings = deepcopy(DEDConfig.default_index_settings())
        self.settings(**default_index_settings)

    def document(self, document):
        """
        Extend to register the document in the global document registry
        """
        document = super(Index, self).document(document)
        registry.register_document(document)
        return document

    doc_type = document

    def __str__(self):
        return self._name



from datetime import datetime
import unittest

import django
from django.core.management import call_command
from django.test import TestCase, TransactionTestCase
if django.VERSION < (4, 0):
    from django.utils.translation import ugettext_lazy as _
else:
    from django.utils.translation import gettext_lazy as _
from six import StringIO

from elasticsearch.exceptions import NotFoundError
from elasticsearch_dsl import Index as DSLIndex
from django_elasticsearch_dsl.test import ESTestCase, is_es_online
from tests import ES_MAJOR_VERSION

from .documents import (
    ad_index,
    AdDocument,
    car_index,
    CarDocument,
    CarWithPrepareDocument,
    ArticleDocument,
    ArticleWithSlugAsIdDocument,
    index_settings
)
from .models import Car, Manufacturer, Ad, Category, Article, COUNTRIES


@unittest.skipUnless(is_es_online(), 'Elasticsearch is offline')
class IntegrationTestCase(ESTestCase, TransactionTestCase):
    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        self.manufacturer = Manufacturer(
            name="Peugeot", created=datetime(1900, 10, 9, 0, 0),
            country_code="FR", logo='logo.jpg'
        )
        self.manufacturer.save()
        self.car1 = Car(
            name="508", launched=datetime(2010, 9, 9, 0, 0),
            manufacturer=self.manufacturer
        )

        self.car1.save()
        self.car2 = Car(
            name="208", launched=datetime(2010, 10, 9, 0, 0),
            manufacturer=self.manufacturer
        )
        self.car2.save()
        self.category1 = Category(
            title="Category 1", slug="category-1", icon="icon.jpeg"
        )
        self.category1.save()
        self.car2.categories.add(self.category1)
        self.car2.save()

        self.car3 = Car(name="308", launched=datetime(2010, 11, 9, 0, 0))
        self.car3.save()
        self.category2 = Category(title="Category 2", slug="category-2")
        self.category2.save()
        self.car3.categories.add(self.category1, self.category2)
        self.car3.save()

        self.ad1 = Ad(
            title=_("Ad number 1"), url="www.ad1.com",
            description="My super ad description 1",
            car=self.car1
        )
        self.ad1.save()
        self.ad2 = Ad(
            title="Ad number 2", url="www.ad2.com",
            description="My super ad descriptio 2",
            car=self.car1
        )
        self.ad2.save()
        self.car1.save()

    def test_get_doc_with_relationships(self):
        s = CarDocument.search().query("match", name=self.car2.name)
        result = s.execute()
        self.assertEqual(len(result), 1)
        car2_doc = result[0]
        self.assertEqual(car2_doc.ads, [])
        self.assertEqual(car2_doc.name, self.car2.name)
        self.assertEqual(int(car2_doc.meta.id), self.car2.pk)
        self.assertEqual(car2_doc.launched, self.car2.launched)
        self.assertEqual(car2_doc.manufacturer.name,
                         self.car2.manufacturer.name)
        self.assertEqual(car2_doc.manufacturer.country,
                         COUNTRIES[self.manufacturer.country_code])

        s = CarDocument.search().query("match", name=self.car3.name)
        result = s.execute()
        car3_doc = result[0]
        self.assertEqual(car3_doc.manufacturer, {})
        self.assertEqual(car3_doc.name, self.car3.name)
        self.assertEqual(int(car3_doc.meta.id), self.car3.pk)

    def test_get_doc_with_reverse_relationships(self):
        s = CarDocument.search().query("match", name=self.car1.name)
        result = s.execute()
        self.assertEqual(len(result), 1)
        car1_doc = result[0]
        self.assertEqual(car1_doc.ads, [
            {
                'title': self.ad1.title,
                'description': self.ad1.description,
                'pk': self.ad1.pk,
            },
            {
                'title': self.ad2.title,
                'description': self.ad2.description,
                'pk': self.ad2.pk,
            },
        ])
        self.assertEqual(car1_doc.name, self.car1.name)
        self.assertEqual(int(car1_doc.meta.id), self.car1.pk)

    def test_get_doc_with_many_to_many_relationships(self):
        s = CarDocument.search().query("match", name=self.car3.name)
        result = s.execute()
        self.assertEqual(len(result), 1)
        car1_doc = result[0]
        self.assertEqual(car1_doc.categories, [
            {
                'title': self.category1.title,
                'slug': self.category1.slug,
                'icon': self.category1.icon.url,
            },
            {
                'title': self.category2.title,
                'slug': self.category2.slug,
                'icon': '',
            }
        ])

    def test_doc_to_dict(self):
        s = CarDocument.search().query("match", name=self.car2.name)
        result = s.execute()
        self.assertEqual(len(result), 1)
        car2_doc = result[0]
        self.assertEqual(car2_doc.to_dict(), {
            'type': self.car2.type,
            'launched': self.car2.launched,
            'name': self.car2.name,
            'manufacturer': {
                'name': self.manufacturer.name,
                'country': COUNTRIES[self.manufacturer.country_code],
            },
            'categories': [{
                'title': self.category1.title,
                'slug': self.category1.slug,
                'icon': self.category1.icon.url,
            }]
        })

        s = CarDocument.search().query("match", name=self.car3.name)
        result = s.execute()
        self.assertEqual(len(result), 1)
        car3_doc = result[0]
        self.assertEqual(car3_doc.to_dict(), {
            'type': self.car3.type,
            'launched': self.car3.launched,
            'name': self.car3.name,
            'categories': [
                {
                    'title': self.category1.title,
                    'slug': self.category1.slug,
                    'icon': self.category1.icon.url,
                },
                {
                    'title': self.category2.title,
                    'slug': self.category2.slug,
                    'icon': '',
                }
            ]
        })

    def test_index_to_dict(self):
        self.maxDiff = None
        index_dict = car_index.to_dict()
        text_type = 'string' if ES_MAJOR_VERSION == 2 else 'text'

        test_index = DSLIndex('test_index').settings(**index_settings)
        test_index.document(CarDocument)

        index_dict = test_index.to_dict()

        self.assertEqual(index_dict['settings'], {
            'number_of_shards': 1,
            'number_of_replicas': 0,
            'analysis': {
                'analyzer': {
                    'html_strip': {
                        'tokenizer': 'standard',
                        'filter': ['lowercase',
                                   'stop', 'snowball'],
                        'type': 'custom',
                        'char_filter': ['html_strip']
                    }
                }
            }
        })
        self.assertEqual(index_dict['mappings'], {
                'properties': {
                    'ads': {
                        'type': 'nested',
                        'properties': {
                            'description': {
                                'type': text_type, 'analyzer':
                                'html_strip'
                            },
                            'pk': {'type': 'integer'},
                            'title': {'type': text_type}
                        },
                    },
                    'categories': {
                        'type': 'nested',
                        'properties': {
                            'title': {'type': text_type},
                            'slug': {'type': text_type},
                            'icon': {'type': text_type}
                        },
                    },
                    'manufacturer': {
                        'type': 'object',
                        'properties': {
                            'country': {'type': text_type},
                            'name': {'type': text_type}
                        },
                    },
                    'name': {'type': text_type},
                    'launched': {'type': 'date'},
                    'type': {'type': text_type}
                }
        })

    def test_related_docs_are_updated(self):
        # test foreignkey relation
        self.manufacturer.name = 'Citroen'
        self.manufacturer.save()

        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(car2_doc.manufacturer.name, 'Citroen')
        self.assertEqual(len(car2_doc.ads), 0)

        ad3 = Ad.objects.create(
            title=_("Ad number 3"), url="www.ad3.com",
            description="My super ad description 3",
            car=self.car2
        )
        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(len(car2_doc.ads), 1)
        ad3.delete()
        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(len(car2_doc.ads), 0)

        self.manufacturer.delete()
        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(car2_doc.manufacturer, {})

    def test_m2m_related_docs_are_updated(self):
        # test m2m add
        category = Category(
            title="Category", slug="category", icon="icon.jpeg"
        )
        category.save()
        self.car2.categories.add(category)
        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(len(car2_doc.categories), 2)

        # test m2m deletion
        self.car2.categories.remove(category)
        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(len(car2_doc.categories), 1)

        self.category1.car_set.clear()
        s = CarDocument.search().query("match", name=self.car2.name)
        car2_doc = s.execute()[0]
        self.assertEqual(len(car2_doc.categories), 0)

    def test_related_docs_with_prepare_are_updated(self):
        s = CarWithPrepareDocument.search().query("match", name=self.car2.name)
        self.assertEqual(s.execute()[0].manufacturer.name, 'Peugeot')
        self.assertEqual(s.execute()[0].manufacturer_short.name, 'Peugeot')

        self.manufacturer.name = 'Citroen'
        self.manufacturer.save()
        s = CarWithPrepareDocument.search().query("match", name=self.car2.name)
        self.assertEqual(s.execute()[0].manufacturer.name, 'Citroen')
        self.assertEqual(s.execute()[0].manufacturer_short.name, 'Citroen')

        self.manufacturer.delete()
        s = CarWithPrepareDocument.search().query("match", name=self.car2.name)
        self.assertEqual(s.execute()[0].manufacturer, {})

    def test_delete_create_populate_commands(self):
        out = StringIO()
        self.assertTrue(ad_index.exists())
        self.assertTrue(car_index.exists())

        call_command('search_index', action='delete',
                     force=True, stdout=out, models=['tests.ad'])
        self.assertFalse(ad_index.exists())
        self.assertTrue(car_index.exists())

        call_command('search_index', action='create',
                     models=['tests.ad'], stdout=out)
        self.assertTrue(ad_index.exists())
        call_command('search_index', action='populate',
                     models=['tests.ad'], stdout=out)
        result = AdDocument().search().execute()
        self.assertEqual(len(result), 2)

    def test_rebuild_command(self):
        out = StringIO()
        result = AdDocument().search().execute()
        self.assertEqual(len(result), 2)

        Ad(title="Ad title 3").save()

        call_command('search_index', action='populate',
                     force=True, stdout=out, models=['tests.ad'])
        result = AdDocument().search().execute()
        self.assertEqual(len(result), 3)

    def test_filter_queryset(self):
        Ad(title="Nothing that match",  car=self.car1).save()

        qs = AdDocument().search().query(
            'match', title="Ad number 2").filter_queryset(Ad.objects)
        self.assertEqual(qs.count(), 2)
        self.assertEqual(list(qs), [self.ad2, self.ad1])

        qs = AdDocument().search().query(
            'match', title="Ad number 2"
        ).filter_queryset(Ad.objects.filter(url="www.ad2.com"))
        self.assertEqual(qs.count(), 1)
        self.assertEqual(list(qs), [self.ad2])

        with self.assertRaisesMessage(TypeError, 'Unexpected queryset model'):
            AdDocument().search().query(
                'match', title="Ad number 2").filter_queryset(Category.objects)

    def test_to_queryset(self):
        Ad(title="Nothing that match",  car=self.car1).save()
        qs = AdDocument().search().query(
            'match', title="Ad number 2").to_queryset()
        self.assertEqual(qs.count(), 2)
        self.assertEqual(list(qs), [self.ad2, self.ad1])

    def test_queryset_iterator_queries(self):
        ad3 = Ad(title="Ad 3",  car=self.car1)
        ad3.save()
        with self.assertNumQueries(1):
            AdDocument().update(Ad.objects.all())

        doc = AdDocument()

        with self.assertNumQueries(1):
            doc.update(Ad.objects.all().order_by('-id'))
            self.assertEqual(
                set(int(instance.meta.id) for instance in
                    doc.search().query('match', title="Ad")),
                set([ad3.pk, self.ad1.pk, self.ad2.pk])
            )

    def test_default_document_id(self):
        obj_id = 12458
        article_slug = "some-article"
        article = Article(
            id=obj_id,
            slug=article_slug,
        )

        # saving should create two documents (in the two indices): one with the
        # Django object's id as the ES doc _id, and the other with the slug
        # as the ES _id
        article.save()

        # assert that the document's id is the id of the Django object
        try:
            es_obj = ArticleDocument.get(id=obj_id)
        except NotFoundError:
            self.fail("document with _id {} not found").format(obj_id)
        self.assertEqual(es_obj.slug, article.slug)

    def test_custom_document_id(self):
        article_slug = "my-very-first-article"
        article = Article(
            slug=article_slug,
        )

        # saving should create two documents (in the two indices): one with the
        # Django object's id as the ES doc _id, and the other with the slug
        # as the ES _id
        article.save()

        # assert that the document's id is its the slug
        try:
            es_obj = ArticleWithSlugAsIdDocument.get(id=article_slug)
        except NotFoundError:
            self.fail(
                "document with _id '{}' not found: "
                "using a custom id is broken".format(article_slug)
            )
        self.assertEqual(es_obj.slug, article.slug)



from django.apps import AppConfig
from django.conf import settings
from django.utils.module_loading import import_string

from elasticsearch_dsl.connections import connections


class DEDConfig(AppConfig):
    name = 'django_elasticsearch_dsl'
    verbose_name = "Django elasticsearch-dsl"
    signal_processor = None

    def ready(self):
        self.module.autodiscover()
        connections.configure(**settings.ELASTICSEARCH_DSL)
        # Setup the signal processor.
        if not self.signal_processor:
            signal_processor_path = getattr(
                settings,
                'ELASTICSEARCH_DSL_SIGNAL_PROCESSOR',
                'django_elasticsearch_dsl.signals.RealTimeSignalProcessor'
            )
            signal_processor_class = import_string(signal_processor_path)
            self.signal_processor = signal_processor_class(connections)

    @classmethod
    def autosync_enabled(cls):
        return getattr(settings, 'ELASTICSEARCH_DSL_AUTOSYNC', True)

    @classmethod
    def default_index_settings(cls):
        return getattr(settings, 'ELASTICSEARCH_DSL_INDEX_SETTINGS', {})

    @classmethod
    def auto_refresh_enabled(cls):
        return getattr(settings, 'ELASTICSEARCH_DSL_AUTO_REFRESH', True)



from types import MethodType

import django
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.fields.files import FieldFile

if django.VERSION < (4, 0):
    from django.utils.encoding import force_text as force_str
else:
    from django.utils.encoding import force_str
from django.utils.functional import Promise
from elasticsearch_dsl.field import (
    Boolean,
    Byte,
    Completion,
    Date,
    Double,
    Field,
    Float,
    GeoPoint,
    GeoShape,
    Integer,
    Ip,
    Long,
    Nested,
    Object,
    ScaledFloat,
    Short,
    Keyword,
    Text,
    SearchAsYouType,
)

from .exceptions import VariableLookupError


class DEDField(Field):
    def __init__(self, attr=None, **kwargs):
        super(DEDField, self).__init__(**kwargs)
        self._path = attr.split('.') if attr else []

    def __setattr__(self, key, value):
        if key == 'get_value_from_instance':
            self.__dict__[key] = value
        else:
            super(DEDField, self).__setattr__(key, value)

    def get_value_from_instance(self, instance, field_value_to_ignore=None):
        """
        Given an model instance to index with ES, return the value that
        should be put into ES for this field.
        """
        if not instance:
            return None

        for attr in self._path:
            try:
                instance = instance[attr]
            except (
                TypeError, AttributeError,
                KeyError, ValueError, IndexError
            ):
                try:
                    instance = getattr(instance, attr)
                except ObjectDoesNotExist:
                    return None
                except (TypeError, AttributeError):
                    try:
                        instance = instance[int(attr)]
                    except (
                        IndexError, ValueError,
                        KeyError, TypeError
                    ):
                        if self._required:
                            raise VariableLookupError(
                                "Failed lookup for key [{}] in "
                                "{!r}".format(attr, instance)
                            )
                        return None

            if isinstance(instance, models.manager.Manager):
                instance = instance.all()
            elif callable(instance):
                instance = instance()
            elif instance is None:
                return None

        if instance == field_value_to_ignore:
            return None

        # convert lazy object like lazy translations to string
        if isinstance(instance, Promise):
            return force_str(instance)

        return instance


class ObjectField(DEDField, Object):
    def _get_inner_field_data(self, obj, field_value_to_ignore=None):
        data = {}

        if hasattr(self, 'properties'):
            for name, field in self.properties.to_dict().items():
                if not isinstance(field, DEDField):
                    continue

                if field._path == []:
                    field._path = [name]

                data[name] = field.get_value_from_instance(
                    obj, field_value_to_ignore
                )
        else:
            doc_instance = self._doc_class()
            for name, field in self._doc_class._doc_type.mapping.properties._params.get(
                'properties', {}).items():  # noqa
                if not isinstance(field, DEDField):
                    continue

                if field._path == []:
                    field._path = [name]

                # This allows for retrieving data from an InnerDoc with prepare_field_name functions.
                prep_func = getattr(doc_instance, 'prepare_%s' % name, None)

                if prep_func:
                    data[name] = prep_func(obj)
                else:
                    data[name] = field.get_value_from_instance(
                        obj, field_value_to_ignore
                    )

        # This allows for ObjectFields to be indexed from dicts with
        # dynamic keys (i.e. keys/fields not defined in 'properties')
        if not data and obj and isinstance(obj, dict):
            data = obj

        return data

    def get_value_from_instance(self, instance, field_value_to_ignore=None):
        objs = super(ObjectField, self).get_value_from_instance(
            instance, field_value_to_ignore
        )

        if objs is None:
            return {}
        try:
            is_iterable = bool(iter(objs))
        except TypeError:
            is_iterable = False

        # While dicts are iterable, they need to be excluded here so
        # their full data is indexed
        if is_iterable and not isinstance(objs, dict):
            return [
                self._get_inner_field_data(obj, field_value_to_ignore)
                for obj in objs if obj != field_value_to_ignore
            ]

        return self._get_inner_field_data(objs, field_value_to_ignore)


def ListField(field):
    """
    This wraps a field so that when get_value_from_instance
    is called, the field's values are iterated over
    """
    original_get_value_from_instance = field.get_value_from_instance

    def get_value_from_instance(self, instance, field_value_to_ignore=None):
        if not original_get_value_from_instance(instance):
            return []
        return [value for value in original_get_value_from_instance(instance)]

    field.get_value_from_instance = MethodType(get_value_from_instance, field)
    return field


class BooleanField(DEDField, Boolean):
    pass


class ByteField(DEDField, Byte):
    pass


class CompletionField(DEDField, Completion):
    pass


class DateField(DEDField, Date):
    pass


class DoubleField(DEDField, Double):
    pass


class FloatField(DEDField, Float):
    pass


class ScaledFloatField(DEDField, ScaledFloat):
    pass


class GeoPointField(DEDField, GeoPoint):
    pass


class GeoShapeField(DEDField, GeoShape):
    pass


class IntegerField(DEDField, Integer):
    pass


class IpField(DEDField, Ip):
    pass


class LongField(DEDField, Long):
    pass


class NestedField(Nested, ObjectField):
    pass


class ShortField(DEDField, Short):
    pass


class KeywordField(DEDField, Keyword):
    pass


class TextField(DEDField, Text):
    pass


class SearchAsYouTypeField(DEDField, SearchAsYouType):
    pass


class FileFieldMixin(object):
    def get_value_from_instance(self, instance, field_value_to_ignore=None):
        _file = super(FileFieldMixin, self).get_value_from_instance(
            instance, field_value_to_ignore)

        if isinstance(_file, FieldFile):
            return _file.url if _file else ''
        return _file if _file else ''


class FileField(FileFieldMixin, DEDField, Text):
    pass


class TimeField(KeywordField):
    def get_value_from_instance(self, instance, field_value_to_ignore=None):
        time = super(TimeField, self).get_value_from_instance(instance,
                                                              field_value_to_ignore)

        if time:
            return time.isoformat()



from mock import Mock
from unittest import TestCase

from django.conf import settings

from django_elasticsearch_dsl import Index
from django_elasticsearch_dsl.registries import DocumentRegistry

from .fixtures import WithFixturesMixin


class DocumentRegistryTestCase(WithFixturesMixin, TestCase):
    def setUp(self):
        self.registry = DocumentRegistry()
        self.index_1 = Index(name='index_1')
        self.index_2 = Index(name='index_2')

        self.doc_a1 = self._generate_doc_mock(self.ModelA, self.index_1)
        self.doc_a2 = self._generate_doc_mock(self.ModelA, self.index_1)
        self.doc_b1 = self._generate_doc_mock(self.ModelB, self.index_2)
        self.doc_c1 = self._generate_doc_mock(self.ModelC, self.index_1)

    def test_empty_registry(self):
        registry = DocumentRegistry()
        self.assertEqual(registry._indices, {})
        self.assertEqual(registry._models, {})

    def test_register(self):
        self.assertEqual(self.registry._models[self.ModelA],
                         set([self.doc_a1, self.doc_a2]))
        self.assertEqual(self.registry._models[self.ModelB],
                         set([self.doc_b1]))

        self.assertEqual(self.registry._indices[self.index_1],
                         set([self.doc_a1, self.doc_a2, self.doc_c1]))
        self.assertEqual(self.registry._indices[self.index_2],
                         set([self.doc_b1]))

    def test_get_models(self):
        self.assertEqual(self.registry.get_models(),
                         set([self.ModelA, self.ModelB, self.ModelC]))

    def test_get_documents(self):
        self.assertEqual(self.registry.get_documents(),
                         set([self.doc_a1, self.doc_a2,
                              self.doc_b1, self.doc_c1]))

    def test_get_documents_by_model(self):
        self.assertEqual(self.registry.get_documents([self.ModelA]),
                         set([self.doc_a1, self.doc_a2]))

    def test_get_documents_by_unregister_model(self):
        ModelC = Mock()
        self.assertFalse(self.registry.get_documents([ModelC]))

    def test_get_indices(self):
        self.assertEqual(self.registry.get_indices(),
                         set([self.index_1, self.index_2]))

    def test_get_indices_by_model(self):
        self.assertEqual(self.registry.get_indices([self.ModelA]),
                         set([self.index_1]))

    def test_get_indices_by_unregister_model(self):
        ModelC = Mock()
        self.assertFalse(self.registry.get_indices([ModelC]))

    def test_update_instance(self):
        doc_a3 = self._generate_doc_mock(
            self.ModelA, self.index_1, _ignore_signals=True
        )

        instance = self.ModelA()
        self.registry.update(instance)

        self.assertFalse(doc_a3.update.called)
        self.assertFalse(self.doc_b1.update.called)
        self.doc_a1.update.assert_called_once_with(instance)
        self.doc_a2.update.assert_called_once_with(instance)

    def test_update_related_instances(self):
        doc_d1 = self._generate_doc_mock(
            self.ModelD, self.index_1,
            _related_models=[self.ModelE, self.ModelB]
        )
        doc_d2 = self._generate_doc_mock(
            self.ModelD, self.index_1, _related_models=[self.ModelE]
        )

        instance_e = self.ModelE()
        instance_b = self.ModelB()
        related_instance = self.ModelD()

        doc_d2.get_instances_from_related.return_value = related_instance
        doc_d1.get_instances_from_related.return_value = related_instance
        self.registry.update_related(instance_e)

        doc_d1.get_instances_from_related.assert_called_once_with(instance_e)
        doc_d1.update.assert_called_once_with(related_instance)
        doc_d2.get_instances_from_related.assert_called_once_with(instance_e)
        doc_d2.update.assert_called_once_with(related_instance)

        doc_d1.get_instances_from_related.reset_mock()
        doc_d1.update.reset_mock()
        doc_d2.get_instances_from_related.reset_mock()
        doc_d2.update.reset_mock()

        self.registry.update_related(instance_b)
        doc_d1.get_instances_from_related.assert_called_once_with(instance_b)
        doc_d1.update.assert_called_once_with(related_instance)
        doc_d2.get_instances_from_related.assert_not_called()
        doc_d2.update.assert_not_called()

    def test_update_related_instances_not_defined(self):
        doc_d1 = self._generate_doc_mock(_model=self.ModelD, index=self.index_1,
                                         _related_models=[self.ModelE])

        instance = self.ModelE()

        doc_d1.get_instances_from_related.return_value = None
        self.registry.update_related(instance)

        doc_d1.get_instances_from_related.assert_called_once_with(instance)
        doc_d1.update.assert_not_called()

    def test_delete_instance(self):
        doc_a3 = self._generate_doc_mock(
            self.ModelA, self.index_1, _ignore_signals=True
        )

        instance = self.ModelA()
        self.registry.delete(instance)

        self.assertFalse(doc_a3.update.called)
        self.assertFalse(self.doc_b1.update.called)
        self.doc_a1.update.assert_called_once_with(instance, action='delete')
        self.doc_a2.update.assert_called_once_with(instance, action='delete')

    def test_autosync(self):
        settings.ELASTICSEARCH_DSL_AUTOSYNC = False

        instance = self.ModelA()
        self.registry.update(instance)
        self.assertFalse(self.doc_a1.update.called)

        settings.ELASTICSEARCH_DSL_AUTOSYNC = True



from __future__ import unicode_literals, absolute_import
from datetime import datetime

from elasticsearch_dsl import connections
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from six.moves import input
from ...registries import registry


class Command(BaseCommand):
    help = 'Manage elasticsearch index.'

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.es_conn = connections.get_connection()

    def add_arguments(self, parser):
        parser.add_argument(
            '--models',
            metavar='app[.model]',
            type=str,
            nargs='*',
            help="Specify the model or app to be updated in elasticsearch"
        )
        parser.add_argument(
            '--create',
            action='store_const',
            dest='action',
            const='create',
            help="Create the indices in elasticsearch"
        )
        parser.add_argument(
            '--populate',
            action='store_const',
            dest='action',
            const='populate',
            help="Populate elasticsearch indices with models data"
        )
        parser.add_argument(
            '--delete',
            action='store_const',
            dest='action',
            const='delete',
            help="Delete the indices in elasticsearch"
        )
        parser.add_argument(
            '--rebuild',
            action='store_const',
            dest='action',
            const='rebuild',
            help="Delete the indices and then recreate and populate them"
        )
        parser.add_argument(
            '-f',
            action='store_true',
            dest='force',
            help="Force operations without asking"
        )
        parser.add_argument(
            '--parallel',
            action='store_true',
            dest='parallel',
            help='Run populate/rebuild update multi threaded'
        )
        parser.add_argument(
            '--no-parallel',
            action='store_false',
            dest='parallel',
            help='Run populate/rebuild update single threaded'
        )
        parser.add_argument(
            '--use-alias',
            action='store_true',
            dest='use_alias',
            help='Use alias with indices'
        )
        parser.add_argument(
            '--use-alias-keep-index',
            action='store_true',
            dest='use_alias_keep_index',
            help="""
                Do not delete replaced indices when used with '--rebuild' and
                '--use-alias' args
            """
        )
        parser.set_defaults(parallel=getattr(settings, 'ELASTICSEARCH_DSL_PARALLEL', False))
        parser.add_argument(
            '--refresh',
            action='store_true',
            dest='refresh',
            default=None,
            help='Refresh indices after populate/rebuild'
        )
        parser.add_argument(
            '--no-count',
            action='store_false',
            default=True,
            dest='count',
            help='Do not include a total count in the summary log line'
        )

    def _get_models(self, args):
        """
        Get Models from registry that match the --models args
        """
        if args:
            models = []
            for arg in args:
                arg = arg.lower()
                match_found = False

                for model in registry.get_models():
                    if model._meta.app_label == arg:
                        models.append(model)
                        match_found = True
                    elif '{}.{}'.format(
                        model._meta.app_label.lower(),
                        model._meta.model_name.lower()
                    ) == arg:
                        models.append(model)
                        match_found = True

                if not match_found:
                    raise CommandError("No model or app named {}".format(arg))
        else:
            models = registry.get_models()

        return set(models)

    def _create(self, models, aliases, options):
        for index in registry.get_indices(models):
            alias_exists = index._name in aliases
            if not alias_exists:
                self.stdout.write("Creating index '{}'".format(index._name))
                index.create()
            elif options['action'] == 'create':
                self.stdout.write(
                    "'{}' already exists as an alias. Run '--delete' with"
                    " '--use-alias' arg to delete indices pointed at the "
                    "alias to make index name available.".format(index._name)
                )

    def _populate(self, models, options):
        parallel = options['parallel']
        for doc in registry.get_documents(models):
            self.stdout.write("Indexing {} '{}' objects {}".format(
                doc().get_queryset().count() if options['count'] else "all",
                doc.django.model.__name__,
                "(parallel)" if parallel else "")
            )
            qs = doc().get_indexing_queryset()
            doc().update(qs, parallel=parallel, refresh=options['refresh'])

    def _get_alias_indices(self, alias):
        alias_indices = self.es_conn.indices.get_alias(name=alias)
        return list(alias_indices.keys())

    def _delete_alias_indices(self, alias):
        alias_indices = self._get_alias_indices(alias)
        alias_delete_actions = [
            {"remove_index": {"index": index}} for index in alias_indices
        ]
        self.es_conn.indices.update_aliases(actions=alias_delete_actions)
        for index in alias_indices:
            self.stdout.write("Deleted index '{}'".format(index))

    def _delete(self, models, aliases, options):
        index_names = [index._name for index in registry.get_indices(models)]

        if not options['force']:
            response = input(
                "Are you sure you want to delete "
                "the '{}' indices? [y/N]: ".format(", ".join(index_names)))
            if response.lower() != 'y':
                self.stdout.write('Aborted')
                return False

        if options['use_alias']:
            for index in index_names:
                alias_exists = index in aliases
                if alias_exists:
                    self._delete_alias_indices(index)
                elif self.es_conn.indices.exists(index=index):
                    self.stdout.write(
                        "'{}' refers to an index, not an alias. Run "
                        "'--delete' without '--use-alias' arg to delete "
                        "index.".format(index)
                    )
                    return False
        else:
            for index in registry.get_indices(models):
                alias_exists = index._name in aliases
                if not alias_exists:
                    self.stdout.write("Deleting index '{}'".format(index._name))
                    index.delete(ignore=404)
                elif options['action'] == 'rebuild':
                    self._delete_alias_indices(index._name)
                elif options['action'] == 'delete':
                    self.stdout.write(
                        "'{}' refers to an alias, not an index. Run "
                        "'--delete' with '--use-alias' arg to delete indices "
                        "pointed at the alias.".format(index._name)
                    )
                    return False

        return True

    def _update_alias(self, alias, new_index, alias_exists, options):
        alias_actions = [{"add": {"alias": alias, "index": new_index}}]

        delete_existing_index = False
        if not alias_exists and self.es_conn.indices.exists(index=alias):
            # Elasticsearch will return an error if an index already
            # exists with the desired alias name. Therefore, we need to
            # delete that index.
            delete_existing_index = True
            alias_actions.append({"remove_index": {"index": alias}})

        old_indices = []
        alias_delete_actions = []
        if alias_exists:
            # Elasticsearch will return an error if we search for
            # indices by alias but the alias doesn't exist. Therefore,
            # we want to be sure the alias exists.
            old_indices = self._get_alias_indices(alias)
            alias_actions.append(
                {"remove": {"alias": alias, "indices": old_indices}}
            )
            alias_delete_actions = [
                {"remove_index": {"index": index}} for index in old_indices
            ]

        self.es_conn.indices.update_aliases(actions=alias_actions)
        if delete_existing_index:
             self.stdout.write("Deleted index '{}'".format(alias))

        self.stdout.write(
            "Added alias '{}' to index '{}'".format(alias, new_index)
        )

        if old_indices:
            for index in old_indices:
                self.stdout.write(
                    "Removed alias '{}' from index '{}'".format(alias, index)
                )

            if alias_delete_actions and not options['use_alias_keep_index']:
                self.es_conn.indices.update_aliases(
                    actions=alias_delete_actions
                )
                for index in old_indices:
                    self.stdout.write("Deleted index '{}'".format(index))

    def _rebuild(self, models, aliases, options):
        if (not options['use_alias']
            and not self._delete(models, aliases, options)):
            return

        if options['use_alias']:
            alias_index_pairs = []
            index_suffix = "-" + datetime.now().strftime("%Y%m%d%H%M%S%f")
            for index in registry.get_indices(models):
                # The alias takes the original index name value. The
                # index name sent to Elasticsearch will be the alias
                # plus the suffix from above. In addition, the index
                # name needs to be limited to 255 characters, of which
                # 21 will always be taken by the suffix, leaving 234
                # characters from the original index name value.
                new_index = index._name[:234] + index_suffix
                alias_index_pairs.append(
                    {'alias': index._name, 'index': new_index}
                )
                index._name = new_index

        self._create(models, aliases, options)
        self._populate(models, options)

        if options['use_alias']:
            for alias_index_pair in alias_index_pairs:
                alias = alias_index_pair['alias']
                alias_exists = alias in aliases
                self._update_alias(
                    alias, alias_index_pair['index'], alias_exists, options
                )

    def handle(self, *args, **options):
        if not options['action']:
            raise CommandError(
                "No action specified. Must be one of"
                " '--create','--populate', '--delete' or '--rebuild' ."
            )

        action = options['action']
        models = self._get_models(options['models'])

        # We need to know if and which aliases exist to mitigate naming
        # conflicts with indices, therefore this is needed regardless
        # of using the '--use-alias' arg.
        aliases = []
        for index in self.es_conn.indices.get_alias().values():
            aliases += index['aliases'].keys()

        if action == 'create':
            self._create(models, aliases, options)
        elif action == 'populate':
            self._populate(models, options)
        elif action == 'delete':
            self._delete(models, aliases, options)
        elif action == 'rebuild':
            self._rebuild(models, aliases, options)
        else:
            raise CommandError(
                "Invalid action. Must be one of"
                " '--create','--populate', '--delete' or '--rebuild' ."
            )



class DjangoElasticsearchDslError(Exception):
    pass


class VariableLookupError(DjangoElasticsearchDslError):
    pass


class RedeclaredFieldError(DjangoElasticsearchDslError):
    pass


class ModelFieldNotMappedError(DjangoElasticsearchDslError):
    pass



from unittest import TestCase
from mock import patch

from django.conf import settings

from django_elasticsearch_dsl.indices import Index
from django_elasticsearch_dsl.registries import DocumentRegistry

from .fixtures import WithFixturesMixin


class IndexTestCase(WithFixturesMixin, TestCase):
    def setUp(self):
        self.registry = DocumentRegistry()

    def test_documents_add_to_register(self):
        registry = self.registry
        with patch('django_elasticsearch_dsl.indices.registry', new=registry):
            index = Index('test')
            doc_a1 = self._generate_doc_mock(self.ModelA)
            doc_a2 = self._generate_doc_mock(self.ModelA)
            index.document(doc_a1)
            docs = list(registry.get_documents())
            self.assertEqual(len(docs), 1)
            self.assertIs(docs[0], doc_a1)

            index.document(doc_a2)
            docs = registry.get_documents()
            self.assertEqual(docs, set([doc_a1, doc_a2]))

    def test__str__(self):
        index = Index('test')
        self.assertEqual(index.__str__(), 'test')

    def test__init__(self):
        settings.ELASTICSEARCH_DSL_INDEX_SETTINGS = {
            'number_of_replicas': 0,
            'number_of_shards': 2,
        }

        index = Index('test')
        self.assertEqual(index._settings, {
            'number_of_replicas': 0,
            'number_of_shards': 2,
        })

        settings.ELASTICSEARCH_DSL_INDEX_SETTINGS = {}



# encoding: utf-8
"""
A convenient way to attach django-elasticsearch-dsl to Django's signals and
cause things to index.
"""

from __future__ import absolute_import

from django.db import models
from django.apps import apps
from django.dispatch import Signal
from .registries import registry
from django.core.exceptions import ObjectDoesNotExist
from importlib import import_module
# Sent after document indexing is completed
post_index = Signal()

class BaseSignalProcessor(object):
    """Base signal processor.

    By default, does nothing with signals but provides underlying
    functionality.
    """

    def __init__(self, connections):
        self.connections = connections
        self.setup()

    def setup(self):
        """Set up.

        A hook for setting up anything necessary for
        ``handle_save/handle_delete`` to be executed.

        Default behavior is to do nothing (``pass``).
        """
        # Do nothing.

    def teardown(self):
        """Tear-down.

        A hook for tearing down anything necessary for
        ``handle_save/handle_delete`` to no longer be executed.

        Default behavior is to do nothing (``pass``).
        """
        # Do nothing.

    def handle_m2m_changed(self, sender, instance, action, **kwargs):
        if action in ('post_add', 'post_remove', 'post_clear'):
            self.handle_save(sender, instance)
        elif action in ('pre_remove', 'pre_clear'):
            self.handle_pre_delete(sender, instance)

    def handle_save(self, sender, instance, **kwargs):
        """Handle save.

        Given an individual model instance, update the object in the index.
        Update the related objects either.
        """
        registry.update(instance)
        registry.update_related(instance)

    def handle_pre_delete(self, sender, instance, **kwargs):
        """Handle removing of instance object from related models instance.
        We need to do this before the real delete otherwise the relation
        doesn't exists anymore and we can't get the related models instance.
        """
        registry.delete_related(instance)

    def handle_delete(self, sender, instance, **kwargs):
        """Handle delete.

        Given an individual model instance, delete the object from index.
        """
        registry.delete(instance, raise_on_error=False)


class RealTimeSignalProcessor(BaseSignalProcessor):
    """Real-time signal processor.

    Allows for observing when saves/deletes fire and automatically updates the
    search engine appropriately.
    """

    def setup(self):
        # Listen to all model saves.
        models.signals.post_save.connect(self.handle_save)
        models.signals.post_delete.connect(self.handle_delete)

        # Use to manage related objects update
        models.signals.m2m_changed.connect(self.handle_m2m_changed)
        models.signals.pre_delete.connect(self.handle_pre_delete)

    def teardown(self):
        # Listen to all model saves.
        models.signals.post_save.disconnect(self.handle_save)
        models.signals.post_delete.disconnect(self.handle_delete)
        models.signals.m2m_changed.disconnect(self.handle_m2m_changed)
        models.signals.pre_delete.disconnect(self.handle_pre_delete)

try:
    from celery import shared_task
except ImportError:
    pass
else:
    class CelerySignalProcessor(RealTimeSignalProcessor):
        """Celery signal processor.

        Allows automatic updates on the index as delayed background tasks using
        Celery.

        NB: We cannot process deletes as background tasks.
        By the time the Celery worker would pick up the delete job, the
        model instance would already deleted. We can get around this by
        setting Celery to use `pickle` and sending the object to the worker,
        but using `pickle` opens the application up to security concerns.
        """

        def handle_save(self, sender, instance, **kwargs):
            """Handle save with a Celery task.

            Given an individual model instance, update the object in the index.
            Update the related objects either.
            """
            pk = instance.pk
            app_label = instance._meta.app_label
            model_name = instance.__class__.__name__

            self.registry_update_task.delay(pk, app_label, model_name)
            self.registry_update_related_task.delay(pk, app_label, model_name)

        def handle_pre_delete(self, sender, instance, **kwargs):
            """Handle removing of instance object from related models instance.
            We need to do this before the real delete otherwise the relation
            doesn't exists anymore and we can't get the related models instance.
            """
            self.prepare_registry_delete_related_task(instance)

        def handle_delete(self, sender, instance, **kwargs):
            """Handle delete.

            Given an individual model instance, delete the object from index.
            """
            self.prepare_registry_delete_task(instance)

        def prepare_registry_delete_related_task(self, instance):
            """
            Select its related instance before this instance was deleted.
            And pass that to celery.
            """
            action = 'index'
            for doc in registry._get_related_doc(instance):
                doc_instance = doc(related_instance_to_ignore=instance)
                try:
                    related = doc_instance.get_instances_from_related(instance)
                except ObjectDoesNotExist:
                    related = None
                if related is not None:
                    doc_instance.update(related)
                    if isinstance(related, models.Model):
                        object_list = [related]
                    else:
                        object_list = related
                    bulk_data = list(doc_instance._get_actions(object_list, action)),
                    self.registry_delete_task.delay(doc_instance.__class__.__name__, bulk_data)

        @shared_task()
        def registry_delete_task(doc_label, data):
            """
            Handle the bulk delete data on the registry as a Celery task.
            The different implementations used are due to the difference between delete and update operations. 
            The update operation can re-read the updated data from the database to ensure eventual consistency, 
            but the delete needs to be processed before the database record is deleted to obtain the associated data.
            """
            doc_instance = import_module(doc_label)
            parallel = True
            doc_instance._bulk(bulk_data, parallel=parallel)

        def prepare_registry_delete_task(self, instance):
            """
            Get the prepare did before database record deleted.
            """
            action = 'delete'
            for doc in registry._get_related_doc(instance):
                doc_instance = doc(related_instance_to_ignore=instance)
                try:
                    related = doc_instance.get_instances_from_related(instance)
                except ObjectDoesNotExist:
                    related = None
                if related is not None:
                    doc_instance.update(related)
                    if isinstance(related, models.Model):
                        object_list = [related]
                    else:
                        object_list = related
                    bulk_data = list(doc_instance.get_actions(object_list, action)),
                    self.registry_delete_task.delay(doc_instance.__class__.__name__, bulk_data)

        @shared_task()
        def registry_update_task(pk, app_label, model_name):
            """Handle the update on the registry as a Celery task."""
            try:
                model = apps.get_model(app_label, model_name)
            except LookupError:
                pass
            else:
                registry.update(
                    model.objects.get(pk=pk)
                )

        @shared_task()
        def registry_update_related_task(pk, app_label, model_name):
            """Handle the related update on the registry as a Celery task."""
            try:
                model = apps.get_model(app_label, model_name)
            except LookupError:
                pass
            else:
                registry.update_related(
                    model.objects.get(pk=pk)
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = '8.0'

if sys.argv[-1] == 'publish':
    try:
        import wheel
        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='django-elasticsearch-dsl',
    version=version,
    python_requires=">=3.8",
    description="""Wrapper around elasticsearch-dsl-py for django models""",
    long_description=readme + '\n\n' + history,
    author='Sabricot',
    url='https://github.com/sabricot/django-elasticsearch-dsl',
    packages=[
        'django_elasticsearch_dsl',
    ],
    include_package_data=True,
    install_requires=[
        'elasticsearch-dsl>=8.9.0,<9.0.0',
        'six',
    ],
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='django elasticsearch elasticsearch-dsl',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    extras_require={
        'celery':  ["celery>=4.1.0"],
    }
)



import django

from django.utils.module_loading import autodiscover_modules

from .documents import Document  # noqa
from .indices import Index  # noqa
from .fields import *  # noqa

__version__ = '7.1.1'


def autodiscover():
    autodiscover_modules('documents')





from django.db.models import Case, When
from django.db.models.fields import IntegerField

from elasticsearch_dsl import Search as DSLSearch


class Search(DSLSearch):
    def __init__(self, **kwargs):
        self._model = kwargs.pop('model', None)
        super(Search, self).__init__(**kwargs)

    def _clone(self):
        s = super(Search, self)._clone()
        s._model = self._model
        return s

    def filter_queryset(self, queryset, keep_search_order=True):
        """
        Filter an existing django queryset using the elasticsearch result.
        It costs a query to the sql db.
        """
        s = self
        if s._model is not queryset.model:
            raise TypeError(
                'Unexpected queryset model '
                '(should be: %s, got: %s)' % (s._model, queryset.model)
            )

        # Do not query again if the es result is already cached
        if not hasattr(self, '_response'):
            # We only need the meta fields with the models ids
            s = self.source(excludes=['*'])
            s = s.execute()

        pks = [result.meta.id for result in s]
        queryset = queryset.filter(pk__in=pks)

        if keep_search_order:
            preserved_order = Case(
                *[When(pk=pk, then=pos) for pos, pk in enumerate(pks)],
                output_field=IntegerField()
            )
            queryset = queryset.order_by(preserved_order)

        return queryset

    def _get_queryset(self):
        """
        Return a django queryset that will be filtered by to_queryset method.
        """
        return self._model._default_manager.all()

    def to_queryset(self, keep_order=True):
        """
        Return a django queryset from the elasticsearch result.
        It costs a query to the sql db.
        """
        qs = self._get_queryset()
        return self.filter_queryset(qs, keep_order)



import json
from unittest import TestCase
from unittest import SkipTest


import django
from django.db import models

if django.VERSION < (4, 0):
    from django.utils.translation import ugettext_lazy as _
else:
    from django.utils.translation import gettext_lazy as _
from elasticsearch_dsl import GeoPoint, InnerDoc
from mock import patch, Mock

from django_elasticsearch_dsl import fields
from django_elasticsearch_dsl.documents import DocType
from django_elasticsearch_dsl.exceptions import (ModelFieldNotMappedError,
                                                 RedeclaredFieldError)
from django_elasticsearch_dsl.registries import registry
from tests import ES_MAJOR_VERSION

from .models import Article


class Car(models.Model):
    name = models.CharField(max_length=255)
    price = models.FloatField()
    not_indexed = models.TextField()
    manufacturer = models.ForeignKey(
        'Manufacturer', null=True, on_delete=models.SET_NULL
    )

    class Meta:
        app_label = 'car'

    def type(self):
        return "break"


class Manufacturer(models.Model):
    name = models.CharField(max_length=255)

    class Meta:
        app_label = 'car'


@registry.register_document
class CarDocument(DocType):
    color = fields.TextField()
    type = fields.TextField()

    def prepare_color(self, instance):
        return "blue"

    class Meta:
        doc_type = 'car_document'

    class Django:
        fields = ['name', 'price']
        model = Car
        related_models = [Manufacturer]

    class Index:
        name = 'car_index'
        doc_type = 'car_document'


class BaseDocTypeTestCase(object):
    TARGET_PROCESSOR = None

    @classmethod
    def setUpClass(cls):
        from django.conf import settings
        if cls.TARGET_PROCESSOR != settings.ELASTICSEARCH_DSL_SIGNAL_PROCESSOR:
            raise SkipTest(
                "Skipped because {} is required, not {}".format(
                    cls.TARGET_PROCESSOR, settings.ELASTICSEARCH_DSL_SIGNAL_PROCESSOR
                )
            )
        super(BaseDocTypeTestCase,cls).setUpClass()

    def test_model_class_added(self):
        self.assertEqual(CarDocument.django.model, Car)

    def test_ignore_signal_default(self):
        self.assertFalse(CarDocument.django.ignore_signals)

    def test_auto_refresh_default(self):
        self.assertTrue(CarDocument.django.auto_refresh)

    def test_ignore_signal_added(self):
        @registry.register_document
        class CarDocument2(DocType):
            class Django:
                model = Car
                ignore_signals = True

        self.assertTrue(CarDocument2.django.ignore_signals)

    def test_auto_refresh_added(self):
        @registry.register_document
        class CarDocument2(DocType):
            class Django:
                model = Car
                auto_refresh = False

        self.assertFalse(CarDocument2.django.auto_refresh)

    def test_queryset_pagination_added(self):
        @registry.register_document
        class CarDocument2(DocType):
            class Django:
                model = Car
                queryset_pagination = 120

        self.assertIsNone(CarDocument.django.queryset_pagination)
        self.assertEqual(CarDocument2.django.queryset_pagination, 120)

    def test_fields_populated(self):
        mapping = CarDocument._doc_type.mapping
        self.assertEqual(
            set(mapping.properties.properties.to_dict().keys()),
            set(['color', 'name', 'price', 'type'])
        )

    def test_related_models_added(self):
        related_models = CarDocument.django.related_models
        self.assertEqual([Manufacturer], related_models)

    def test_duplicate_field_names_not_allowed(self):
        with self.assertRaises(RedeclaredFieldError):
            @registry.register_document
            class CarDocument(DocType):
                color = fields.TextField()
                name = fields.TextField()

                class Django:
                    fields = ['name']
                    model = Car

    def test_to_field(self):
        doc = DocType()
        nameField = doc.to_field('name', Car._meta.get_field('name'))
        self.assertIsInstance(nameField, fields.TextField)
        self.assertEqual(nameField._path, ['name'])

    def test_to_field_with_unknown_field(self):
        doc = DocType()
        with self.assertRaises(ModelFieldNotMappedError):
            doc.to_field('manufacturer', Car._meta.get_field('manufacturer'))

    def test_mapping(self):
        text_type = 'string' if ES_MAJOR_VERSION == 2 else 'text'

        self.assertEqual(
            CarDocument._doc_type.mapping.to_dict(), {
                'properties': {
                    'name': {
                        'type': text_type
                    },
                    'color': {
                        'type': text_type
                    },
                    'type': {
                        'type': text_type
                    },
                    'price': {
                        'type': 'double'
                    }
                }
            }
        )

    def test_get_queryset(self):
        qs = CarDocument().get_queryset()
        self.assertIsInstance(qs, models.QuerySet)
        self.assertEqual(qs.model, Car)

    def test_prepare(self):
        car = Car(name="Type 57", price=5400000.0, not_indexed="not_indexex")
        doc = CarDocument()
        prepared_data = doc.prepare(car)
        self.assertEqual(
            prepared_data, {
                'color': doc.prepare_color(None),
                'type': car.type(),
                'name': car.name,
                'price': car.price
            }
        )

    def test_innerdoc_prepare(self):
        class ManufacturerInnerDoc(InnerDoc):
            name = fields.TextField()
            location = fields.TextField()

            def prepare_location(self, instance):
                return "USA"

        @registry.register_document
        class CarDocumentWithInnerDoc(DocType):
            manufacturer = fields.ObjectField(doc_class=ManufacturerInnerDoc)

            class Django:
                model = Car
                fields = ['name', 'price']

            class Index:
                name = 'car_index'

        manufacturer = Manufacturer(
            name="Bugatti",
        )

        car = Car(name="Type 57", price=5400000.0, manufacturer=manufacturer)
        doc = CarDocumentWithInnerDoc()
        prepared_data = doc.prepare(car)
        self.assertEqual(
            prepared_data, {
                'name': car.name,
                'price': car.price,
                'manufacturer': {
                    'name': car.manufacturer.name,
                    'location': ManufacturerInnerDoc().prepare_location(manufacturer)
                }
            }
        )

    def test_prepare_ignore_dsl_base_field(self):
        @registry.register_document
        class CarDocumentDSlBaseField(DocType):
            position = GeoPoint()

            class Django:
                model = Car
                fields = ['name', 'price']

            class Index:
                name = 'car_index'

        car = Car(name="Type 57", price=5400000.0, not_indexed="not_indexex")
        doc = CarDocumentDSlBaseField()
        prepared_data = doc.prepare(car)
        self.assertEqual(
            prepared_data, {
                'name': car.name,
                'price': car.price
            }
        )

    def test_model_instance_update(self):
        doc = CarDocument()
        car = Car(name="Type 57", price=5400000.0,
                  not_indexed="not_indexex", pk=51)
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update(car)
            actions = [{
                '_id': car.pk,
                '_op_type': 'index',
                '_source': {
                    'name': car.name,
                    'price': car.price,
                    'type': car.type(),
                    'color': doc.prepare_color(None),
                },
                '_index': 'car_index',
            }]
            self.assertEqual(1, mock.call_count)
            self.assertEqual(
                actions, list(mock.call_args_list[0][1]['actions'])
            )
            self.assertTrue(mock.call_args_list[0][1]['refresh'])
            self.assertEqual(
                doc._index.connection, mock.call_args_list[0][1]['client']
            )

    def test_model_instance_iterable_update(self):
        doc = CarDocument()
        car = Car(name="Type 57", price=5400000.0,
                  not_indexed="not_indexex", pk=51)
        car2 = Car(name=_("Type 42"), price=50000.0,
                   not_indexed="not_indexex", pk=31)
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update([car, car2], action='update')
            actions = [{
                '_id': car.pk,
                '_op_type': 'update',
                '_source': {
                    'name': car.name,
                    'price': car.price,
                    'type': car.type(),
                    'color': doc.prepare_color(None),
                },
                '_index': 'car_index',
            },
                {
                    '_id': car2.pk,
                    '_op_type': 'update',
                    '_source': {
                        'name': car2.name,
                        'price': car2.price,
                        'type': car2.type(),
                        'color': doc.prepare_color(None),
                    },
                    '_index': 'car_index'
                }]
            self.assertEqual(1, mock.call_count)
            self.assertEqual(
                actions, list(mock.call_args_list[0][1]['actions'])
            )
            self.assertTrue(mock.call_args_list[0][1]['refresh'])
            self.assertEqual(
                doc._index.connection, mock.call_args_list[0][1]['client']
            )

    def test_model_instance_update_no_refresh(self):
        doc = CarDocument()
        doc.django.auto_refresh = False
        car = Car()
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update(car)
            self.assertNotIn('refresh', mock.call_args_list[0][1])

    def test_model_instance_update_refresh_true(self):
        doc = CarDocument()
        doc.django.auto_refresh = False
        car = Car()
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update(car, refresh=True)
            self.assertEqual(
                mock.call_args_list[0][1]['refresh'], True
            )

    def test_model_instance_update_refresh_wait_for(self):
        doc = CarDocument()
        doc.django.auto_refresh = False
        car = Car()
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update(car, refresh='wait_for')
            self.assertEqual(
                mock.call_args_list[0][1]['refresh'], 'wait_for'
            )

    def test_model_instance_update_auto_refresh_wait_for(self):
        doc = CarDocument()
        doc.django.auto_refresh = 'wait_for'
        car = Car()
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update(car)
            self.assertEqual(
                mock.call_args_list[0][1]['refresh'], 'wait_for'
            )

    def test_model_instance_update_refresh_overrides_auto_refresh(self):
        doc = CarDocument()
        doc.django.auto_refresh = True
        car = Car()
        with patch('django_elasticsearch_dsl.documents.bulk') as mock:
            doc.update(car, refresh=False)
            self.assertEqual(
                mock.call_args_list[0][1]['refresh'], False
            )

    def test_model_instance_iterable_update_with_pagination(self):
        class CarDocument2(DocType):
            class Django:
                model = Car
                queryset_pagination = 2

        doc = CarDocument()
        car1 = Car()
        car2 = Car()
        car3 = Car()

        bulk = "django_elasticsearch_dsl.documents.bulk"
        parallel_bulk = "django_elasticsearch_dsl.documents.parallel_bulk"
        with patch(bulk) as mock_bulk, patch(parallel_bulk) as mock_parallel_bulk:
            doc.update([car1, car2, car3])
            self.assertEqual(
                3, len(list(mock_bulk.call_args_list[0][1]['actions']))
            )
            self.assertEqual(mock_bulk.call_count, 1, "bulk is called")
            self.assertEqual(mock_parallel_bulk.call_count, 0,
                             "parallel bulk is not called")

    def test_model_instance_iterable_update_with_parallel(self):
        class CarDocument2(DocType):
            class Django:
                model = Car

        doc = CarDocument()
        car1 = Car()
        car2 = Car()
        car3 = Car()
        bulk = "django_elasticsearch_dsl.documents.bulk"
        parallel_bulk = "django_elasticsearch_dsl.documents.parallel_bulk"
        with patch(bulk) as mock_bulk, patch(parallel_bulk) as mock_parallel_bulk:
            doc.update([car1, car2, car3], parallel=True)
            self.assertEqual(mock_bulk.call_count, 0, "bulk is not called")
            self.assertEqual(mock_parallel_bulk.call_count, 1, "parallel bulk is called")

    def test_init_prepare_correct(self):
        """Does init_prepare() run and collect the right preparation functions?"""

        d = CarDocument()
        self.assertEqual(len(d._prepared_fields), 4)

        expect = {
            'color': ("<class 'django_elasticsearch_dsl.fields.TextField'>",
                      ("<class 'method'>", "<type 'instancemethod'>")),  # py3, py2
            'type': ("<class 'django_elasticsearch_dsl.fields.TextField'>",
                     ("<class 'functools.partial'>", "<type 'functools.partial'>")),
            'name': ("<class 'django_elasticsearch_dsl.fields.TextField'>",
                     ("<class 'functools.partial'>", "<type 'functools.partial'>")),
            'price': ("<class 'django_elasticsearch_dsl.fields.DoubleField'>",
                      ("<class 'functools.partial'>", "<type 'functools.partial'>")),
        }

        for name, field, prep in d._prepared_fields:
            e = expect[name]
            self.assertEqual(str(type(field)), e[0], 'field type should be copied over')
            self.assertTrue('__call__' in dir(prep), 'prep function should be callable')
            self.assertTrue(str(type(prep)) in e[1],
                            'prep function is correct partial or method')

    def test_init_prepare_results(self):
        """Are the results from init_prepare() actually used in prepare()?"""
        d = CarDocument()

        car = Car()
        setattr(car, 'name', "Tusla")
        setattr(car, 'price', 340123.21)
        setattr(car, 'color', "polka-dots")  # Overwritten by prepare function
        setattr(car, 'pk', 4701)  # Ignored, not in document
        setattr(car, 'type', "imaginary")

        self.assertEqual(d.prepare(car),
                         {'color': 'blue', 'type': 'imaginary', 'name': 'Tusla',
                          'price': 340123.21})

        m = Mock()
        # This will blow up should we access _fields and try to iterate over it.
        # Since init_prepare compiles a list of prepare functions, while
        # preparing no access to _fields should happen
        with patch.object(CarDocument, '_fields', 33):
            d.prepare(m)
        self.assertEqual(sorted([tuple(x) for x in m.method_calls], key=lambda _: _[0]),
                         [('name', (), {}), ('price', (), {}), ('type', (), {})]
                         )

    # Mock the elasticsearch connection because we need to execute the bulk so that the generator
    # got iterated and generate_id called.
    # If we mock the bulk in django_elasticsearch_dsl.document
    # the actual bulk will be never called and the test will fail
    @patch('elasticsearch_dsl.connections.Elasticsearch.bulk')
    def test_default_generate_id_is_called(self, _):
        article = Article(
            id=124594,
            slug='some-article',
        )

        @registry.register_document
        class ArticleDocument(DocType):
            class Django:
                model = Article
                fields = [
                    'slug',
                ]

            class Index:
                name = 'test_articles'
                settings = {
                    'number_of_shards': 1,
                    'number_of_replicas': 0,
                }

        with patch.object(ArticleDocument, 'generate_id',
                          return_value=article.id) as patched_method:
            d = ArticleDocument()
            d.update(article)
            patched_method.assert_called()

    @patch('elasticsearch_dsl.connections.Elasticsearch.bulk')
    def test_custom_generate_id_is_called(self, mock_bulk):
        article = Article(
            id=54218,
            slug='some-article-2',
        )

        @registry.register_document
        class ArticleDocument(DocType):
            class Django:
                model = Article
                fields = [
                    'slug',
                ]

            class Index:
                name = 'test_articles'

            @classmethod
            def generate_id(cls, article):
                return article.slug

        d = ArticleDocument()
        d.update(article)

        # Get the data from the elasticsearch low level API because
        # The generator get executed there.
        data = json.loads(mock_bulk.call_args[1]['operations'][1])
        assert data['slug'] == article.slug

    @patch('elasticsearch_dsl.connections.Elasticsearch.bulk')
    def test_should_index_object_is_called(self, mock_bulk):
        doc = CarDocument()
        car1 = Car()
        car2 = Car()
        car3 = Car()
        should_index_object = ("django_elasticsearch_dsl.documents."
                               "DocType.should_index_object")
        with patch(should_index_object) as mock_should_index_object:
            doc.update([car1, car2, car3])
            # As we are indexing 3 object, it should be called 3 times
            self.assertEqual(mock_should_index_object.call_count, 3,
                             "should_index_object is called")

    @patch('elasticsearch_dsl.connections.Elasticsearch.bulk')
    def test_should_index_object_working_perfectly(self, mock_bulk):
        article1 = Article(slug='article1')
        article2 = Article(slug='article2')

        @registry.register_document
        class ArticleDocument(DocType):
            class Django:
                model = Article
                fields = [
                    'slug',
                ]

            class Index:
                name = 'test_articles'

            def should_index_object(self, obj):
                # Article with slug article1 should not be indexed
                if obj.slug == "article2":
                    return False
                return True

        d = ArticleDocument()
        d.update([article1, article2])
        operations = mock_bulk.call_args[1]['operations']
        slugs = [
            json.loads(operation)['slug'] for operation in operations
            if 'slug' in json.loads(operation)
        ]
        self.assertTrue(article1.slug in slugs)
        self.assertTrue(article2.slug not in slugs)

class RealTimeDocTypeTestCase(BaseDocTypeTestCase, TestCase):
    TARGET_PROCESSOR = 'django_elasticsearch_dsl.signals.RealTimeSignalProcessor'


class CeleryDocTypeTestCase(BaseDocTypeTestCase, TestCase):
    TARGET_PROCESSOR = 'django_elasticsearch_dsl.signals.CelerySignalProcessor'



import re

from django.test.utils import captured_stderr
from elasticsearch_dsl.connections import connections

from ..registries import registry


def is_es_online(connection_alias='default'):
    with captured_stderr():
        es = connections.get_connection(connection_alias)
        return es.ping()


class ESTestCase(object):
    _index_suffixe = '_ded_test'

    def setUp(self):
        for doc in registry.get_documents():
            doc._index._name += self._index_suffixe

        for index in registry.get_indices():
            index._name += self._index_suffixe
            index.delete(ignore=[404, 400])
            index.create()

        super(ESTestCase, self).setUp()

    def tearDown(self):
        pattern = re.compile(self._index_suffixe + '$')

        for index in registry.get_indices():
            index.delete(ignore=[404, 400])
            index._name = pattern.sub('', index._name)

        for doc in registry.get_documents():
            doc._index._name = pattern.sub('', doc._index._name)

        super(ESTestCase, self).tearDown()
